/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef _3DROMED_KERNEL_H
#define _3DROMED_KERNEL_H

#include <rg3DRom.h>
#include <vector>
#include <qwidget.h>

// FIXME: Spelling mistake.
#define sel_verticies sel_vertices
#define sel_verticies_count sel_vertices_count

#define KED_MODE_NONE		0
#define KED_MODE_CAMERA_TRAN	1
#define KED_MODE_CAMERA_YROT	2
#define KED_MODE_CAMERA_XROT	3
#define KED_MODE_CAMERA_ZROT	4
#define KED_MODE_CAMERA_SCALE	5
#define KED_MODE_SEL_VERT	6
#define KED_MODE_ADD_VERT	7
#define KED_MODE_SEL_TRI	8
#define KED_MODE_ADD_TRI	9
#define KED_MODE_SEL_BONE	10
#define KED_MODE_ADD_BONE	11
#define KED_MODE_SEL_BRUSH	12
#define KED_MODE_ADD_BRUSH	13
#define KED_MODE_SCALE_BRUSH	14
#define KED_MODE_ROTATE_BRUSH	15
#define KED_MODE_ROTATE 	16
#define KED_MODE_SCALE		17


#define KRN_SCALE_BRUSH_XYZ	0
#define KRN_SCALE_BRUSH_X	1
#define KRN_SCALE_BRUSH_Y	2
#define KRN_SCALE_BRUSH_Z	3

#define KRN_PIXMAP_ICON_RED_H	0

#define KRN_MIN_BONE_LENGTH 4.0f

class GLView;
class frmMain;
class frmView;
class frmGroup;
class frmTexPage;
class frmFrame;
class frmSequence;
class QPixmap;
class QApplication;

extern std::vector <QPixmap> gbl_pixmaps;
extern QApplication *gbl_app;

class ViewData
{
public:
	bool axis_aligned;
	bool edit_window;
	bool show_cross_hair;
	bool show_grid;
	bool show_rulers;
	bool use_wireframe;
	bool use_lighting;
	bool use_texture;

	float ortho_width;
	float ortho_height;

	int mouse_down_x;
	int mouse_down_y;
	int mouse_move_x;
	int mouse_move_y;
	bool mouse_moved;
	bool do_cam_scale;
	bool do_cam_rotate;
	bool do_vertex_move;
	bool show_vertex_sel_box;
	bool do_sel_box_move;
	
	float small_grid_res;
	float medium_grid_res;
	float large_grid_res;
	float world_pixel_length;

	rgGLLight light[4];

	frmView *frm;

	ViewData();
};

class SelectionBox
{
public:
	void set_box( const rgVector mdp, const rgVector mcp);
	void set_box( const float mdx, const float mdy, const float mcx,
		const float mcy, const GLView *g);
	void set_offset( const rgVector off) { m_offset = off;}
	void set_mouse_down( const rgVector mp) { m_mouse_down = mp;}
	bool mouse_in_box( const rgVector mp) const {
		if( rgPlanarPointInConvexFace( m_v, 4, mp)) return true;
		return false;
	}
	rgVector box_center() const {
		return (0.25f * (m_v[0] + m_v[1] + m_v[2] + m_v[3]));
	}
	const rgVector *vertices() const { return m_v;}
	void transform( const rgMatrix &m) {
		m_v[0] = m.pointMul( m_v[0]);
		m_v[1] = m.pointMul( m_v[1]);
		m_v[2] = m.pointMul( m_v[2]);
		m_v[3] = m.pointMul( m_v[3]);
	}
	rgVector offset() const { return m_offset;}
	rgVector mouse_down() const { return m_mouse_down;}
	void render() const { render( rgMatrix());}
	void render( const rgMatrix &m) const;

	void set_selection_center( const rgVector cen){ m_sel_center = cen;}
	rgVector selection_center() const { return m_sel_center;}

	void set_transformation_matrix( const rgMatrix &m) {
		m_transformation = m;
	}

	rgMatrix transformation_matrix() const { return m_transformation;}

private:
	rgVector m_sel_center;
	rgVector m_v[4];
	rgVector m_mouse_down;	// Mouse down point (in world space)
	rgVector m_offset;
	rgMatrix m_transformation;
};

class Model
{
public:
	QString fname;
	bool snap_to_grid;
	bool modified;

	int scale_brush_mode;
	rgCoordSystem temp_brush_csys;
	bool translate_brush;

	int sel_vertex;
	int sel_triangle;

	int sel_vertices_count;
	int *sel_vertices;

	int sel_bone;
	int sel_joint;

	SelectionBox sel_box;

	// Total number of verticies that are part of the triangle being
	// built.
	int tri_build_count;
	int tri_build_vid[3];
	rgVector tri_build_mp;

	bool show_bone_cursor;
	rgVector bone_build_v[4];
	bool rotate_bone;
	bool twist_bone;
	bool translate_bone;
	bool do_bone_bind;
	bool do_bone_mount;
	int bone_bind_slave;
	rgCoordSystem sel_bone_rcs;

	rgVector selection_center_point;
	rgMatrix tool_matrix;	// Tool Editing matrix

	sk3DRomModel mdl;

	frmGroup *frmGrp;
	frmTexPage *frmTex;
	frmFrame *frmSkel;
	frmSequence *frmSeq;
	std::vector <frmView *> view;

	void create_four_view();
	frmView *create_view_window( QWidget *parent);
	void delete_all_view_windows();
	void destroy_view( frmView *v);
	void set_caption_for_all_views( const char *s);

	void add_vertex( frmView *p_view, const rgVector v);

	Model();
	~Model();

	void clear();
	void clear_verticies_selection();
	void show_all_view_windows() const;
	void update_all_views() const;

	rgVector get_selection_box_center() const {
		return sel_box.box_center();
	}

	void set_selecting_view( frmView *);
	frmView *selecting_view() const { return m_selecting_view;}
private:
	// The view that the user most recently made a selection in.
	frmView *m_selecting_view;
};

void vfn_initializeGL( GLView *g);
void vfn_destructorGL( GLView *g);
void vfn_paintGL( GLView *g);
void vfn_resizeGL( int w, int h, GLView *g);
void vfn_mouse_down( QMouseEvent *e, GLView *g);
void vfn_mouse_move( QMouseEvent *e, GLView *g);
void vfn_mouse_up( QMouseEvent *e, GLView *g);
void vfn_key_down( QKeyEvent *e, GLView *g);

void krn_create_main_tool_buttons( frmMain *frm);
Model *krn_create_new_model( std::vector <Model *> &mdl);
void krn_delete_model( Model *m);
int krn_get_prev_edit_mode();
int krn_get_edit_mode();

int krn_select_vertex( Model &mdl, rgVector p, const rgVector n,
	const float max_dis);
void krn_select_vertices( Model &mdl, const rgVector n);

frmView *krn_create_model_view_window( Model &mdl);
void krn_call_open_model();
void krn_call_new_model();
void krn_cascade_windows();
void krn_tile_windows();
rgVector krn_snap_to_grid( ViewData &vd, const rgVector v);
int krn_load_system_pixmaps();
void krn_add_model_to_editor( Model *mdl);
void krn_remove_model_from_editor( const Model *mdl);
void krn_set_selection_center_point( ViewData &vd);
void krn_set_transformed_selection_box( ViewData &vd);
void krn_hlp_gen_rotation_matrix( const rgVector focus, const rgVector axis,
	const rgVector a, const rgVector b, const int ang_lock, rgMatrix &m);
rgMatrix krn_gen_rotation_matrix( const rgVector p_init, const rgVector p_fini,
	const rgVector center, const rgVector axis, const int ang_lock);
rgMatrix krn_gen_scale_matrix( const float length, const rgVector win_init,
	const rgVector win_fini, const rgVector center, const float scale);

void krn_do_brush_mouse_down( QMouseEvent *e, GLView *g, ViewData &vd);
void krn_do_brush_mouse_move( QMouseEvent *e, GLView *g, ViewData &vd);
void krn_print_location( const rgVector p);

#endif
