/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "kernel.h"
#include "frmmain.h"
#include "frmview.h"
#include "glView.h"
#include "kernel_grid.h"
#include <qcursor.h>
#include <qpopupmenu.h>
#include <rg3DRom_csg.h>

void krn_do_brush_mouse_down( QMouseEvent *e, GLView *g, ViewData &vd)
{
	sk3DRomModel &mdl = vd.frm->mdl->mdl;
	Model &kmdl = *(vd.frm->mdl);

	rgVector mwc[] = {	// Mouse World Coordinates
		rgVector( vd.mouse_down_x, vd.mouse_down_y, 0),
		rgVector( e->x(), e->y(), 0)};
	rggl_ortho_window_to_world( g->width(),
		g->height(), mwc, 2, &(g->vmat.icamM));
	const rgVector mdp( mwc[0]);	// Mouse Down Point (World coord)
	//const rgVector mcp( mwc[1]);	// Mouse Current Point (World coord)

	if( mdl.m_brush_cursor)
		kmdl.temp_brush_csys = mdl.m_brush_cursor->coord_system();

	switch( krn_get_edit_mode()) {
	case KED_MODE_SEL_BRUSH:
		kmdl.translate_brush = false;
		if( mdl.m_brush_cursor) {
			const rgVector a( rgMakePlanar(
				mdl.m_brush_cursor->position(), g->view_normal()));
			const rgVector b( rgMakePlanar( mdp, g->view_normal()));
			if( (b - a).length() <= vd.world_pixel_length * 40.0f) {
				kmdl.translate_brush = true;
				kmdl.update_all_views();
			}
		}
		if( !(kmdl.translate_brush)) {
			mdl.m_brush_cursor = mdl.m_csg->brush(
				mdl.m_csg->select_brush( mdp,
					g->view_normal(),
					vd.world_pixel_length * 100.0f));
			kmdl.update_all_views();
		}
		break;

	case KED_MODE_ADD_BRUSH:
		mdl.m_brush_cursor = new sk3DRomCSGCube( &mdl);
		mdl.m_brush_cursor->set_position( mdp);
		mdl.m_brush_cursor->scale( 64, 64, 64);
		mdl.m_csg->add_brush( mdl.m_brush_cursor);
		kmdl.update_all_views();
		break;
	}
}

void krn_do_brush_mouse_move( QMouseEvent *e, GLView *g, ViewData &vd)
{
	sk3DRomModel &mdl = vd.frm->mdl->mdl;
	Model &kmdl = *(vd.frm->mdl);

	rgVector mwc[] = {	// Mouse World Coordinates
		rgVector( vd.mouse_down_x, vd.mouse_down_y, 0),
		rgVector( e->x(), e->y(), 0)};
	rggl_ortho_window_to_world( g->width(),
		g->height(), mwc, 2, &(g->vmat.icamM));
	const rgVector mdp( mwc[0]);	// Mouse Down Point (World coord)
	const rgVector mcp( mwc[1]);	// Mouse Current Point (World coord)

	switch( krn_get_edit_mode()) {
	case KED_MODE_SEL_BRUSH:
		if( !(mdl.m_brush_cursor)) break;
		if( kmdl.translate_brush) {
			const rgMatrix inv( g->vmat.icamM);
			rgGLViewMatrix m( g->vmat);
			m.do_view_translate( g->width(), g->height(),
				rgVector( vd.mouse_down_x, vd.mouse_down_y, 0),
				rgVector( e->x(), e->y(), 0), 0, 0, false);
			mdl.m_brush_cursor->set_coord_system(
				kmdl.temp_brush_csys);
			mdl.m_brush_cursor->translate( inv * (m.camM));
			kmdl.update_all_views();
		}
		break;

	case KED_MODE_ROTATE_BRUSH:
		if( mdl.m_brush_cursor) {
			const rgMatrix inv( g->vmat.icamM);
			rgGLViewMatrix m( g->vmat);
			m.do_view_3way_rotate( g->width(), g->height(),
				rgVector( vd.mouse_down_x, vd.mouse_down_y, 0),
				rgVector( e->x(), e->y(), 0), 0, 0, false);
			mdl.m_brush_cursor->set_coord_system(
				kmdl.temp_brush_csys);
			mdl.m_brush_cursor->rotate( inv * (m.camM));
			kmdl.update_all_views();
		}
		break;

	case KED_MODE_SCALE_BRUSH:
		if( mdl.m_brush_cursor) {
			rgGLViewMatrix m;
			m.do_view_scale( g->height(),
				rgVector( vd.mouse_down_x, vd.mouse_down_y, 0),
				rgVector( e->x(), e->y(), 0),
				2.0f, 0, 0, false);
			const float t = (m.camM * rgVector( 1, 0, 0)).length();
			mdl.m_brush_cursor->set_coord_system(
				kmdl.temp_brush_csys);
			switch( kmdl.scale_brush_mode) {
			case KRN_SCALE_BRUSH_X:
				mdl.m_brush_cursor->scale( t, 1, 1); break;
			case KRN_SCALE_BRUSH_Y:
				mdl.m_brush_cursor->scale( 1, t, 1); break;
			case KRN_SCALE_BRUSH_Z:
				mdl.m_brush_cursor->scale( 1, 1, t); break;
			default: mdl.m_brush_cursor->scale( t, t, t); break;
			}
			kmdl.update_all_views();
		}
		break;

	case KED_MODE_ADD_BRUSH:
		mdl.m_brush_cursor->set_position( mcp);
		kmdl.update_all_views();
		break;
	}
}
