/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <qlistbox.h>
#include "frmtexpage.h"
#include "glView.h"
#include "texture_page_support.h"
#include "frmview.h"
#include <qpopupmenu.h>
#include <qcursor.h>

PageViewData::PageViewData( frmTexPage *p_frm)
{
	show_sel_box = false;
	m_sel_face = -1;
	frm = p_frm;
}

void do_paintGL_texture_view( frmTexPage *frm)
{
	glClear( GL_COLOR_BUFFER_BIT);
	glLoadIdentity();

	const int tid = frm->lstTex->currentItem();
	if( tid < 0) return;
	frm->mdl->mdl.texture[tid].bind();
	glBegin( GL_QUADS);
	glTexCoord2i( 0, 1);
	glVertex2i( -1, 1);

	glTexCoord2i( 1, 1);
	glVertex2i( 1, 1);

	glTexCoord2i( 1, 0);
	glVertex2i( 1, -1);

	glTexCoord2i( 0, 0);
	glVertex2i( -1, -1);
	glEnd();
}

void do_resizeGL_texture_view( int w, int h, frmTexPage *frm)
{
	glViewport( 0, 0, w, h);
	glMatrixMode( GL_PROJECTION);
	glLoadIdentity();
	glOrtho( -1, 1, -1, 1, 1, -1);
	glMatrixMode( GL_MODELVIEW);
}

void do_initGL_texture_view( frmTexPage *frm)
{
	glClearColor( 1, 1, 1, 1);
	glEnable( GL_TEXTURE_2D);
}

void do_paintGL_page_view( frmTexPage *frm)
{
	glClear( GL_COLOR_BUFFER_BIT);
	GLView *g = frm->glvPage;

	glMatrixMode( GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho( -1, 1, -1, 1, 1, -1);
	glMatrixMode( GL_MODELVIEW);
	glLoadIdentity();
	GLubyte pattern[1024];
	memset( pattern, 0, 1024);
	GLubyte *p = pattern;
	for( int i = 0; i < 4; i++) {
		for( int j = 0; j < 32; j++)
			*(p + j) = 0xFF * (((j>>1) + ((i>>1) & 1)) & 1);
		p += 32;
	}

	glPolygonMode( GL_FRONT_AND_BACK, GL_FILL);
	glPolygonStipple( pattern);
	glEnable( GL_POLYGON_STIPPLE);
	glBegin( GL_QUADS);
	glColor3f( 0.5, 0.5, 0.5);
	glVertex2f( -1, 1);
	glVertex2f( 1, 1);
	glVertex2f( 1, -1);
	glVertex2f( -1, -1);
	glEnd();
	glColor3f( 1, 1, 1);
	glDisable( GL_POLYGON_STIPPLE);
	//glPolygonStipple( 0);
	glMatrixMode( GL_PROJECTION);
	glPopMatrix();
	glMatrixMode( GL_MODELVIEW);

	loadRGMatrix( g->vmat.camM);

	const int pid = frm->lstPage->currentItem();
	if( pid < 0) return;
	frm->mdl->mdl.render_page( pid);
	PageViewData &pvd = *((PageViewData *)(frm->glvPage->edata));
	sk3DRomModel &mdl = pvd.frm->mdl->mdl;

	switch( krn_get_edit_mode()) {
	case KED_MODE_SEL_TRI:
		if( pvd.sel_face() >= 0) {
			const int fid = pvd.sel_face();
			glEnable( GL_BLEND);
			glBegin( GL_TRIANGLES);
			glColor4f( 0, 1, 1, 0.5f);
			mdl.render_page_face( pid, fid);
			glEnd();
			glDisable( GL_BLEND);

			glPolygonMode( GL_FRONT_AND_BACK, GL_LINE);
			glLineWidth( 5);
			glBegin( GL_TRIANGLES);
			glColor3f( 0, 1, 1);
			mdl.render_page_face( pid, fid);
			glEnd();
			glLineWidth( 1);
			glPolygonMode( GL_FRONT_AND_BACK, GL_FILL);

			glPointSize( 5);
			glBegin( GL_POINTS);
			glColor3f( 1, 0, 0);
			glVertex( ((mdl.page( pid))->face( fid))->v(0));
			glColor3f( 0, 1, 0);
			glVertex( ((mdl.page( pid))->face( fid))->v(1));
			glColor3f( 0, 0, 1);
			glVertex( ((mdl.page( pid))->face( fid))->v(2));
			glEnd();
		}
		break;

	case KED_MODE_SEL_VERT:	// Display the selected vertex
		if( (mdl.page( pid))->vertex_is_selected()) {
			glPointSize( 10);
			glBegin( GL_POINTS);
			switch( (mdl.page( pid))->selected_vertex_vid()) {
			case 0: glColor3f( 1, 0, 0); break;
			case 1: glColor3f( 0, 1, 0); break;
			case 2: glColor3f( 0, 0, 1); break;
			default: glColor3f( 0, 0, 0); break;
			}
			glVertex( (mdl.page( pid))->selected_vertex());
			glEnd();
		}
		break;
	}

	

	
	if( pvd.show_sel_box) {
		pvd.sel_box.render();
		glPointSize( 7);
		glBegin( GL_POINTS);
		glColor3f( 0, 1, 1);
		(frm->mdl->mdl.page( pid))->render_selected_vertices();
		glEnd();
	}
}

void do_resizeGL_page_view( int w, int h, frmTexPage *frm)
{
	glViewport( 0, 0, w, h);
	glMatrixMode( GL_PROJECTION);
	glLoadIdentity();
	glOrtho( -w * 0.5f, w * 0.5f, -h * 0.5f, h * 0.5f, 1, -1);
	glMatrixMode( GL_MODELVIEW);
}

void do_initGL_page_view( frmTexPage *frm)
{
	glClearColor( 0.75, 0.75, 0.75, 1);
	glEnable( GL_TEXTURE_2D);
	frm->glvPage->edata = new PageViewData( frm);
	frm->glvPage->vmat.set_transformation_order( RGGL_VIEW_MATRIX_TRS);
	reset_page_view( frm);

	glFrontFace( GL_CW);
	glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}

void do_destroy_page_view( frmTexPage *frm)
{
	delete ((PageViewData *)(frm->glvPage->edata));
	frm->glvPage->edata = 0;
}

void reset_page_view( frmTexPage *frm)
{
	frm->glvPage->vmat.reset();
	//frm->glvPage->vmat.translate( -0.5f * frm->glvPage->width(),
		//-0.5f * frm->glvPage->height(), 0.0f);
	frm->glvPage->vmat.translate( -0.5f, -0.5f, 0.0f);
	const float t = RG_MIN( frm->glvPage->width(), frm->glvPage->height())
		- 20.0f;
	frm->glvPage->vmat.scale( t, t, 1);
	frm->glvPage->vmat.update();
}

void do_page_popup_menu( PageViewData &pvd, Model &mdl)
{
	sk3DRomTexturePage *p = mdl.mdl.edit_page(
		pvd.frm->lstPage->currentItem());
	if( !p) return;
	if( p->selected_vertex_count() <= 0) {
		for( int i = 100; i < 108; i++)
			pvd.frm->mnuPage->setItemVisible( i, false);
	} else {
		for( int i = 100; i < 108; i++)
			pvd.frm->mnuPage->setItemVisible( i, true);
	}
	(pvd.frm->mnuPage)->popup( QCursor::pos());
}

void do_mouse_down_page_view( QMouseEvent *e, GLView *g)
{
	PageViewData &pvd = *((PageViewData *)g->edata);
	sk3DRomModel &mdl = pvd.frm->mdl->mdl;
	const int pid = pvd.frm->lstPage->currentItem();
	pvd.mouse_down_x = e->x();
	pvd.mouse_down_y = e->y();
	pvd.mouse_moved = false;

	if( Qt::RightButton == e->button()) {
		do_page_popup_menu( pvd, *(pvd.frm->mdl));
		return;
	}

	switch( krn_get_edit_mode()) {
	case KED_MODE_CAMERA_TRAN:
		pvd.do_cam_scale = false;
		if( (e->state() & Qt::ShiftButton)) pvd.do_cam_scale = true;
		break;

	case KED_MODE_ROTATE:
		if( pid < 0) return;
		pvd.sel_box.set_selection_center(
			(mdl.page( pid))->get_center_of_selected_vertices());
		break;

	case KED_MODE_SCALE:
		if( pid < 0) return;
		pvd.sel_box.set_selection_center(
			(mdl.page( pid))->get_center_of_selected_vertices());
		break;

	case KED_MODE_SEL_VERT:
		pvd.move_sel_vertex = false;
		pvd.move_sel_box = false;
		if( (pid >= 0)  && (mdl.page( pid))->vertex_is_selected()) {
			rgVector v[] = {
				rgVector( e->x(), e->y(), 0),
				rgVector( e->x() + 16, e->y(), 0)};
			rggl_ortho_window_to_world( g->width(), g->height(),
				v, 2, &(g->vmat.icamM));
			if( ((mdl.page(pid))->selected_vertex() - v[0]).length()
				<= (v[0] - v[1]).length()) {
				pvd.move_sel_vertex = true;
				return;
			}
		}
		
		if( pvd.show_sel_box) {
			rgVector mp( e->x(), e->y(), 0);
			rggl_ortho_window_to_world( g->width(), g->height(),
				&mp, 1, &(g->vmat.icamM));
			if( pvd.sel_box.mouse_in_box( mp)) {
				pvd.move_sel_box = true;
				pvd.sel_box.set_mouse_down( mp);
				return;
			}
		}

		pvd.clear_sel_box( mdl.edit_page( pid));
		g->updateGL();
		break;

	case KED_MODE_SEL_TRI:
		{
			if(	(pvd.sel_face() >= 0) &&
				(Qt::ShiftButton & e->state())) {
				rgVector mp( e->x(), e->y(), 0);
				rggl_ortho_window_to_world( g->width(), g->height(),
					&mp, 1, &(g->vmat.icamM));
				pvd.sel_face_offset[0] =
					((mdl.page( pid))->face(
						pvd.sel_face()))->v(0) - mp;
				pvd.sel_face_offset[1] =
					((mdl.page( pid))->face(
						pvd.sel_face()))->v(1) - mp;
				pvd.sel_face_offset[2] =
					((mdl.page( pid))->face(
						pvd.sel_face()))->v(2) - mp;
						
				pvd.move_sel_face = true;

			} else {
				pvd.move_sel_face = false;
				rgVector mp( e->x(), e->y(), 0);
				rggl_ortho_window_to_world( g->width(), g->height(),
					&mp, 1, &(g->vmat.icamM));
				if( pvd.select_face( mp) >= 0) {
					pvd.frm->mdl->sel_triangle =
						((mdl.page(pid))->face( pvd.sel_face()))->
							fid();
				}
				pvd.frm->mdl->update_all_views();
			}
		}
		g->updateGL();
		break;
	}
}

void do_mouse_move_page_view( QMouseEvent *e, GLView *g, frmTexPage *frm)
{
	PageViewData &pvd = *((PageViewData *)g->edata);
	sk3DRomModel &mdl = pvd.frm->mdl->mdl;
	const int pid = pvd.frm->lstPage->currentItem();
	pvd.mouse_moved = true;

	rgVector mwc[] = {	// Mouse World Coordinates
		rgVector( pvd.mouse_down_x, pvd.mouse_down_y, 0),
		rgVector( e->x(), e->y(), 0)};
	rggl_ortho_window_to_world( g->width(),
		g->height(), mwc, 2, &(g->vmat.icamM));
	const rgVector mdp = mwc[0];	// Mouse Down Point (World coord)
	const rgVector mcp = mwc[1];	// Mouse Current Point (World coord)

	switch( krn_get_edit_mode()) {
	case KED_MODE_ROTATE:
		{
			int ang_lock = -1;
			if( (Qt::ShiftButton & e->state())) ang_lock = 15;
			pvd.sel_box.set_transformation_matrix(
				krn_gen_rotation_matrix( mdp, mcp,
					pvd.sel_box.selection_center(),
					g->view_normal(),
					ang_lock));
		}
		g->updateGL();
		break;

	case KED_MODE_SCALE:
		pvd.sel_box.set_transformation_matrix(
			krn_gen_scale_matrix( g->height(),
				rgVector( pvd.mouse_down_x, pvd.mouse_down_y, 0),
				rgVector( e->x(), e->y(), 0),
				pvd.sel_box.selection_center(),
				2.0f));
		g->updateGL();
		break;

	case KED_MODE_SEL_TRI:
		if( pvd.move_sel_face) {
			rgVector mp( e->x(), e->y(), 0);
			rggl_ortho_window_to_world( g->width(), g->height(),
				&mp, 1, &(g->vmat.icamM));
			mdl.update_page_face( pid, pvd.sel_face(),
				pvd.sel_face_offset[0] + mp,
				pvd.sel_face_offset[1] + mp,
				pvd.sel_face_offset[2] + mp);
			g->updateGL();
		}
		break;

	case KED_MODE_SEL_VERT:
		if( pvd.move_sel_box) {
			rgVector mp( e->x(), e->y(), 0);
			rggl_ortho_window_to_world( g->width(), g->height(),
				&mp, 1, &(g->vmat.icamM));
			pvd.sel_box.set_offset( mp - pvd.sel_box.mouse_down());
			g->updateGL();

		} else if( pvd.move_sel_vertex && (pid >= 0)) {
			rgVector mp( e->x(), e->y(), 0);
			rggl_ortho_window_to_world( g->width(), g->height(),
				&mp, 1, &(g->vmat.icamM));
			mdl.update_selected_page_vertex( pid, mp);
			g->updateGL();

		} else {
			pvd.sel_box.set_box(
				pvd.mouse_down_x, pvd.mouse_down_y,
				e->x(), e->y(),
				g);
			pvd.show_sel_box = true;
			g->updateGL();
		}
		break;

	case KED_MODE_CAMERA_TRAN:
		if( pvd.do_cam_scale) {
			g->vmat.do_view_scale( g->height(),
				rgVector( pvd.mouse_down_x, pvd.mouse_down_y, 0),
				rgVector( e->x(), e->y(), 0), 2.0f, 0, 0, false);
			g->updateGL();	

		} else {
			g->vmat.do_view_translate( g->width(), g->height(),
				rgVector( pvd.mouse_down_x, pvd.mouse_down_y, 0),
				rgVector( e->x(), e->y(), 0), 0, 0, false);
			g->updateGL();
		}
		break;
	}
}

void do_mouse_up_page_view( QMouseEvent *e, GLView *g)
{
	PageViewData &pvd = *((PageViewData *)g->edata);
	sk3DRomModel &mdl = pvd.frm->mdl->mdl;
	const int pid = pvd.frm->lstPage->currentItem();

	switch( krn_get_edit_mode()) {
	case KED_MODE_ROTATE:
	case KED_MODE_SCALE:
		mdl.transform_selected_page_vertices( pid,
			pvd.sel_box.transformation_matrix());
		pvd.sel_box.transform( pvd.sel_box.transformation_matrix());
		pvd.sel_box.set_transformation_matrix( rgMatrix());
		g->updateGL();
		break;

	case KED_MODE_SEL_VERT:
		if( pvd.move_sel_box) {
			if( pid < 0) return;
			rgMatrix t;
			t.translate( pvd.sel_box.offset());
			mdl.transform_selected_page_vertices( pid, t);
			pvd.sel_box.transform( t);
			pvd.sel_box.set_offset( rgVector());
			pvd.move_sel_box = false;
			g->updateGL();

		} else if( !pvd.mouse_moved) {
			rgVector v[] = {
				rgVector( e->x(), e->y(), 0),
				rgVector( e->x() + 32, e->y(), 0)};
			rggl_ortho_window_to_world( g->width(), g->height(),
				v, 2, &(g->vmat.icamM));
			mdl.select_page_vertex( pid, v[0], (v[0] - v[1]).length());
			g->updateGL();

		} else {
			if( pid < 0) return;
			mdl.select_page_vertices(
				pid,
				g->view_normal(),
				pvd.sel_box.vertices(),
				4);
			if( (mdl.page( pid))->selected_vertex_count() <= 0)
				pvd.show_sel_box = false;
			g->updateGL();
		}
		break;

	case KED_MODE_CAMERA_TRAN:
		if( pvd.do_cam_scale) {
			g->vmat.do_view_scale( g->height(),
				rgVector( pvd.mouse_down_x, pvd.mouse_down_y, 0),
				rgVector( e->x(), e->y(), 0),
				2.0f, 0, 0, true);
			g->updateGL();

		} else {
			g->vmat.do_view_translate( g->width(), g->height(),
				rgVector( pvd.mouse_down_x, pvd.mouse_down_y, 0),
				rgVector( e->x(), e->y(), 0), 0, 0, true);
			g->updateGL();
		}
		break;
	}
}

void hlp_insert_selected_face_to_page(
	QKeyEvent *e, GLView *g, frmTexPage *frm)
{
	const int pid = frm->lstPage->currentItem();
	const int fid = frm->mdl->sel_triangle;
	if( fid < 0) return;
	sk3DRomFace &f = frm->mdl->mdl.f[fid];
	sk3DRomModel &tmp = frm->mdl->mdl;
	if( Qt::ShiftButton & e->state()) {
		rgVector tri[] = {
			tmp.get_tv_vertex( f.vid[0]),
			tmp.get_tv_vertex( f.vid[1]),
			tmp.get_tv_vertex( f.vid[2])};
		const rgVector w0( tri[1] - tri[0]);
		const rgVector w1( tri[2] - tri[1]);
		const rgVector w2( tri[0] - tri[2]);
		float t0 = fabs( w0.dot( rgVector( 0, 1, 0)));
		float t1 = fabs( w1.dot( rgVector( 0, 1, 0)));
		float t2 = fabs( w2.dot( rgVector( 0, 1, 0)));
		rgVector x;
		if( t0 < t1) {
			if( t0 < t2) x = w0;
			else x = w2;
		} else {
			if( t1 < t2) x = w1;
			else x = w2;
		}
		x.normalize();
		if( x.dot( rgVector( 1, 0, 0)) < 0) x = -x;
		const rgVector y( ((w1.cross( w0)).normalize()).cross( x));
		const rgVector des_x( 1, 0, 0);
		const rgVector des_y( 0, 1, 0);
		tri[1] =	x.dot( tri[1] - tri[0]) * des_x +
				y.dot( tri[1] - tri[0]) * des_y;
		tri[2] =	x.dot( tri[2] - tri[0]) * des_x +
				y.dot( tri[2] - tri[0]) * des_y;
		tri[0] = rgVector();
		frm->mdl->mdl.add_tv_face_to_page(
			pid,
			fid,
			tri[0],
			tri[1],
			tri[2]);

	} else {
		rgMatrix m;
		{
			frmView *view = frm->mdl->selecting_view();
			if( view) m = view->glv->vmat.camM;
		}
		rgVector tri[] = {
			m.pointMul( tmp.get_tv_vertex( f.vid[0])),
			m.pointMul( tmp.get_tv_vertex( f.vid[1])),
			m.pointMul( tmp.get_tv_vertex( f.vid[2])),
			rgVector( g->width() >> 1, g->height() >> 1, 0)
		};
		rggl_ortho_world_to_window( g->width(), g->height(), tri, 3);
		rggl_ortho_window_to_world( g->width(), g->height(),
			tri, 4, &(g->vmat.icamM));
		tri[3] = tri[3] - tri[0];
		frm->mdl->mdl.add_tv_face_to_page(
			pid,
			fid,
			tri[0] + tri[3],
			tri[1] + tri[3],
			tri[2] + tri[3]);
	}
	g->updateGL();
}

void do_key_down_page_view( QKeyEvent *e, GLView *g, frmTexPage *frm)
{
	PageViewData &pvd = *((PageViewData *)g->edata);
	const int pid = frm->lstPage->currentItem();
	switch( krn_get_edit_mode()) {
	case KED_MODE_SEL_VERT:
		switch( e->key()) {
		case Qt::Key_F:
			pvd.frm->mdl->mdl.fuse_selected_page_vertices( pid);
			g->updateGL();
			break;
		}
		break;

	case KED_MODE_SEL_TRI:
		switch( e->key()) {
		case Qt::Key_Delete:
			{
				sk3DRomTexturePage *p =
					pvd.frm->mdl->mdl.edit_page( pid);
				if( !p) return;
				p->delete_face( pvd.sel_face());
				pvd.clear_selected_face();
				g->updateGL();
			}
			break;

		case Qt::Key_Insert:
			hlp_insert_selected_face_to_page( e, g, frm);
			break;
		}
		break;
	}
}




